
package com.datastax.devcenter.ide.contentassist

import com.datastax.devcenter.cql.CqlContext
import com.datastax.devcenter.cql.cql.AlterTableStatement
import com.datastax.devcenter.cql.cql.ColumnEntity
import com.datastax.devcenter.cql.cql.ColumnOperation
import com.datastax.devcenter.cql.cql.CqlStatement
import com.datastax.devcenter.cql.cql.CreateIndexStatement
import com.datastax.devcenter.cql.cql.DeleteOp
import com.datastax.devcenter.cql.cql.DeleteStatement
import com.datastax.devcenter.cql.cql.InsertStatement
import com.datastax.devcenter.cql.cql.KeyspaceEntity
import com.datastax.devcenter.cql.cql.SelectClause
import com.datastax.devcenter.cql.cql.SelectStatement
import com.datastax.devcenter.cql.cql.Selector
import com.datastax.devcenter.cql.cql.SelectorFunction
import com.datastax.devcenter.cql.cql.TableEntity
import com.datastax.devcenter.cql.cql.UpdateStatement
import com.datastax.devcenter.cql.services.CqlGrammarAccess
import com.datastax.devcenter.cql.util.ModelUtil
import com.datastax.devcenter.schema.Column
import com.datastax.devcenter.schema.ISchemaElement
import com.datastax.devcenter.schema.Schema
import com.datastax.devcenter.schema.Table
import com.google.inject.Inject
import java.lang.reflect.Method
import java.util.ArrayList
import java.util.Collection
import java.util.HashMap
import java.util.LinkedHashSet
import java.util.List
import org.apache.commons.lang.StringUtils
import org.eclipse.emf.ecore.EObject
import org.eclipse.xtext.AbstractRule
import org.eclipse.xtext.Assignment
import org.eclipse.xtext.GrammarUtil
import org.eclipse.xtext.Keyword
import org.eclipse.xtext.RuleCall
import org.eclipse.xtext.TerminalRule
import org.eclipse.xtext.ide.editor.contentassist.ContentAssistContext
import org.eclipse.xtext.ide.editor.contentassist.ContentAssistEntry
import org.eclipse.xtext.ide.editor.contentassist.IIdeContentProposalAcceptor
import org.eclipse.xtext.ide.editor.contentassist.IdeContentProposalProvider
import org.eclipse.xtext.util.Strings

import static com.datastax.devcenter.cql.util.ModelUtil.getCqlStatement
import static com.datastax.devcenter.cql.util.ModelUtil.getKeyspace
import static com.datastax.devcenter.cql.util.ModelUtil.getTable
import static com.datastax.devcenter.cql.util.ModelUtil.stripName
import com.google.common.base.Optional
import com.datastax.devcenter.cql.BaseCqlContextProvider

@SuppressWarnings(value="all")
class IdeCqlProposalProvider extends IdeContentProposalProvider {
	
 	@Inject
    private BaseCqlContextProvider cqlContextProvider;

	var dispatchers = new HashMap<String, Optional<Method>>();

	override protected def dispatch void createProposals(
		Assignment assignment, ContentAssistContext context,
			IIdeContentProposalAcceptor acceptor) {
		completeAssignment(assignment, context, acceptor)
	}		
	
	def void completeAssignment(
		Assignment assignment, 
		ContentAssistContext contentAssistContext,
		IIdeContentProposalAcceptor acceptor) {
			
		var parserRule = GrammarUtil.containingParserRule(assignment);
		var methodName = "complete" + Strings.toFirstUpper(parserRule.getName()) + "_"
				+ Strings.toFirstUpper(assignment.getFeature());

		var Method found = null;
		
		if(dispatchers.get(methodName) == null){
			for(Method m : class.getDeclaredMethods()){
				if(m.name.equals(methodName)){
					found = m
				}	
			}
			dispatchers.put(methodName, Optional.fromNullable(found))
			
		}
		var m = dispatchers.get(methodName);
		if (m != null && m.isPresent) {
			m.get.invoke(this,
				contentAssistContext.getCurrentModel(),
				assignment,
				contentAssistContext,
				acceptor
			)
		}
	}

	
    public def completeSelectStatement_Table(
        EObject model,
        Assignment assignment,
        ContentAssistContext context,
        IIdeContentProposalAcceptor acceptor) {
        proposeTables(model, assignment, context, acceptor)
    }
   
    def void completeSelector_Column(
        EObject model,
        Assignment assignment,
        ContentAssistContext context,
        IIdeContentProposalAcceptor acceptor) {
        proposeColumns(model, assignment, context, acceptor);
    } 
     
    def void complete_SelectorFunctionArgs(
        EObject model,
        ContentAssistContext context,
        IIdeContentProposalAcceptor acceptor) {
        if (model != null && model instanceof SelectorFunction) {
            var SelectorFunction selectorFunction = model as SelectorFunction;
            /*
             * This only is useful if a table has already been specified,
             * otherwise columns can't be proposed.
             */
            if (selectorFunction.getFunctionName() != null &&
                selectorFunction.getFunctionName().getName() != null) {
                var String name = selectorFunction.getFunctionName().getName();
                if (name.equalsIgnoreCase("toJson") ||
                    name.equalsIgnoreCase("max") ||
                    name.equalsIgnoreCase("min") ||
                    name.equalsIgnoreCase("avg") ||
                    name.equalsIgnoreCase("sum")) {
                    proposeColumns("", "", model, context, acceptor);
                }
            }
        }
    }
    
     def void proposeColumns(
        EObject model,
        Assignment assignment,
        ContentAssistContext context,
        IIdeContentProposalAcceptor acceptor) {
        proposeColumns("", "", model, context, acceptor);
    }

    def void proposeColumns(
        String prefix, String suffix, EObject model,
        ContentAssistContext context,
        IIdeContentProposalAcceptor acceptor) {

        var CqlContext cqlContext = cqlContextProvider.get(model);
        var CqlStatement stmt = getCqlStatement(model);

        if ((stmt instanceof AlterTableStatement)
            && (stmt as AlterTableStatement).getOperation().equalsIgnoreCase("add")) {
            return; // no proposal for ADD
        }

        var index = cqlContext.getStatementIndexMap().getStatementIndex(stmt);
        if ((stmt != null) && (index >= 0)) {
            var KeyspaceEntity keyspaceEntity = getKeyspace(stmt);
            var TableEntity tableEntity = getTable(stmt);
            var Schema schema = cqlContext.getSchema(index);
            var Table tableMetadata = schema.getTable(keyspaceEntity, tableEntity);

            if (tableMetadata != null) {
                var Collection<Column> columns = 
                	new LinkedHashSet<Column>(tableMetadata.getColumns());
                	
                maybeFilterPrimaryKeyColumns(stmt, tableMetadata, columns);

                proposeEntitiesNotInCollection(
                    prefix,
                    suffix,
                    columns,
                    collectUsedColumnNames(model),
                    context,
                    acceptor);
            }
        }
    }
    
    def proposeEntitiesNotInCollection(
        String prefix, String suffix, Collection<? extends ISchemaElement> entities,
        Collection<? extends ISchemaElement> filterCollection,
        ContentAssistContext context,
        IIdeContentProposalAcceptor acceptor) {
        var List<ISchemaElement> proposedEntities = new ArrayList<ISchemaElement>();
        for (ISchemaElement entry : entities) {
            if (!filterCollection.contains(entry)) {
                proposedEntities.add(entry);
            }
        }
        proposeNames(proposedEntities, context, acceptor);
    }
    
     /**
     * Collect column names used earlier in a CqlStatement
     *
     * @param model
     * @return
     */
    def Collection<Column> collectUsedColumnNames(EObject model) {
        var List<Column> proposedColumns = new ArrayList<Column>();
        var CqlStatement stmt = getCqlStatement(model);

        var KeyspaceEntity keyspaceEntity = getKeyspace(stmt);
        var TableEntity tableEntity = getTable(stmt);
        if (tableEntity == null) {
            return proposedColumns;
        }

        var CqlContext cqlContext = cqlContextProvider.get(model);
        var int index = cqlContext.getStatementIndexMap().getStatementIndex(stmt);
        if (index == -1) {
            return proposedColumns;
        }

        var Schema schema = cqlContext.getSchema(index);
        if (schema == null) {
            return proposedColumns;
        }

        var Table cfMeta = schema.getTable(keyspaceEntity, tableEntity);
        if (cfMeta == null) {
            return proposedColumns;
        }

        var List<String> proposedNames = collectProposedColumnNames(stmt);
        for (Column column : cfMeta.getColumns()) {
            if (proposedNames.contains(column.getName())) {
                proposedColumns.add(column);
            }
        }
        return proposedColumns;
    }
    
    def List<String> collectProposedColumnNames(CqlStatement stmt) {
        var List<String> proposed = new ArrayList<String>();
        if (stmt instanceof SelectStatement) {
            var SelectClause sc = stmt.getSelectClause();
            if ((sc != null) && (sc.getSelectors() != null)) {
                for (Selector sel : sc.getSelectors()) {
                    if (sel.getColumn() != null) {
                        proposed.add(stripName(sel.getColumn().getName()));
                    }
                }
            }

        } else if (stmt instanceof InsertStatement) {
            var InsertStatement insert = stmt;
            if ((insert.getStandardInsertPart() != null)
                && (insert.getStandardInsertPart().getColumns() != null)) {
                for (ColumnEntity colName : insert.getStandardInsertPart().getColumns()) {
                    proposed.add(stripName(colName.getName()));
                }
            }

        } else if (stmt instanceof UpdateStatement) {
            for (ColumnOperation colop : stmt.getOperations()) {
                if (colop.getColumn() != null) {
                    proposed.add(stripName(colop.getColumn().getName()));
                }
            }

        } else if (stmt instanceof DeleteStatement) {
            for (DeleteOp delop : stmt.getSelections()) {
                if (delop.getColumn() != null) {
                    proposed.add(stripName(delop.getColumn().getName()));
                }
            }

        } else if (stmt instanceof AlterTableStatement) {
            for (ColumnEntity colName : stmt.getOldColumns()) {
                proposed.add(stripName(colName.getName()));
            }
        } // end of collecting
        return proposed;
    }
    
    def maybeFilterPrimaryKeyColumns(
        CqlStatement stmt,
        Table tableMetadata,
        Collection<Column> columns) {
        if ((stmt instanceof CreateIndexStatement)
            || ((stmt instanceof AlterTableStatement) && (stmt as AlterTableStatement).
            	getOperation().equalsIgnoreCase(
                "drop"))) {
            // Don't include PK columns
            columns.removeAll(tableMetadata.getPrimaryKeys());
        }
    }
	    
    @Inject extension CqlGrammarAccess
	override protected _createProposals(
		RuleCall ruleCall, 
		ContentAssistContext context,
		IIdeContentProposalAcceptor acceptor) {
			
		if (ruleCall.rule instanceof TerminalRule) {
			createDefaultProposal(ruleCall.rule, context, acceptor);
		} else {
			switch (ruleCall.rule) {
				case k_INSERTRule:
					createDefaultProposal(ruleCall.rule, context, acceptor)
				case k_INTORule:
					createDefaultProposal(ruleCall.rule, context, acceptor)
				case k_VALUESRule:
					createDefaultProposal(ruleCall.rule, context, acceptor)
				case k_SELECTRule:
					createDefaultProposal(ruleCall.rule, context, acceptor)
				case k_FROMRule:
					createDefaultProposal(ruleCall.rule, context, acceptor)
				case k_CREATERule:
					createDefaultProposal(ruleCall.rule, context, acceptor)
				case k_TABLERule:
					createDefaultProposal(ruleCall.rule, context, acceptor)
				default:
					super._createProposals(ruleCall, context, acceptor)
			}
		}
	}

	protected def createDefaultProposal(AbstractRule rule, ContentAssistContext context,
		IIdeContentProposalAcceptor acceptor) {
		val keyword = rule.alternatives as Keyword
		if (keyword.value.startsWith(context.prefix)) {
			val entry = new ContentAssistEntry => [
				prefix = context.prefix
				proposal = keyword.value
			]
			acceptor.accept(entry, proposalPriorities.getDefaultPriority(entry))
		}
	}
    
    protected def proposeTables(
        EObject model,
        Assignment assignment,
        ContentAssistContext context,
        IIdeContentProposalAcceptor acceptor) {
        	
        var cqlContext = cqlContextProvider.get(model)
        var stmt = getCqlStatement(model)

        var int index = cqlContext.getStatementIndexMap().getStatementIndex(stmt)
        if ((stmt != null) && (index >= 0)) {
            var keyspaceEntity = ModelUtil.getKeyspace(stmt)
            var schema = cqlContext.getSchema(index)
            var ksMeta = schema.getKeyspace(keyspaceEntity)

            if (ksMeta != null) {
                proposeNames(ksMeta.getTables(), context, acceptor)
            }
        }
    }
    
     def void proposeNames(
        Collection<? extends ISchemaElement> entities,
        ContentAssistContext context,
        IIdeContentProposalAcceptor acceptor) {
        if (entities != null) {
            for (ISchemaElement entity : entities) {
                if (!StringUtils.isEmpty(entity.getName())) {
                    proposeName(entity, context, acceptor);
                }
            }
        }
    }

    // propose single name
    def void proposeName(
        ISchemaElement entity,
        ContentAssistContext context,
        IIdeContentProposalAcceptor acceptor) {
        proposeText(entity.name, entity.name, context, acceptor)
    }
    
    def void proposeText(
    	String enter, 
    	String descr, 
    	ContentAssistContext context,
		IIdeContentProposalAcceptor acceptor) {

		val entry = new ContentAssistEntry => [
			prefix = context.prefix
			description = descr
			proposal = enter
		]
		acceptor.accept(entry, proposalPriorities.getDefaultPriority(entry))
	}
}
